# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""**Run** provides a class to conveniently read and record experiment metrics and artifacts."""

import logging
import os
import re
import json

from collections import defaultdict
from six import raise_from
from inspect import isgeneratorfunction
from types import GeneratorType
from uuid import uuid4

from azureml.exceptions import RunEnvironmentException, UserErrorException
from azureml._logging.chained_identity import ChainedIdentity
from azureml._run_impl.run_base import _RunBase
from azureml._history.utils.constants import OUTPUTS_DIR
from azureml._restclient.constants import BASE_RUN_SOURCE, RUN_ORIGIN, SDK_TARGET
from azureml._restclient.utils import camel_case_transformer

module_logger = logging.getLogger(__name__)


def _check_for_data_container_id(func):
    def wrapped(self, *args, **kwargs):
        if self._container is None:
            raise UserErrorException("{} doesn't have a data container associated with it yet - "
                                     "therefore, the {} cannot upload files, or log file backed metrics.".format(
                                         self, self.__class__.__name__))
        return func(self, *args, **kwargs)
    return wrapped


class Run(_RunBase):
    """The base class for all experiment runs.

    A *run* represents a single trial of an experiment, and allows monitoring and logging.

    .. remarks::

        A *run* represents a single trial of an experiment.  A run is the object used to monitor the
        asynchronous execution of a trial, log metrics and store output of the trial,
        and to analyze results and access artifacts generated by the trial.

        Run is used inside of your experimentation code to log metrics and artifacts to the Run History service.

        Run is used  outside of your experiments monitor progress and to query and analyze
        the metrics and results that were generated.

        Functionality includes:

        *  Storing and retrieving metrics and data
        *  Uploading and downloading files
        *  Using tags as well as the child hierarchy for easy lookup of past runs
        *  Registering stored model files as model's that can be operationalized
        *  Storing, modifying and retrieving properties of a run
        *  Loading the current run from a remote environment with Run.get_context()
        *  Efficiently snapshotting a file or directory for reproducibility

        This class works with the :class:`azureml.core.Experiment` and is used in these scenarios

        #. Creating a run by executing code using :func:`azureml.core.Experiment.submit`
        #. Creating a run interactively in a notebook using :func:`azureml.core.Experiment.start_logging`
        #. Logging metrics and uploading artifacts in your experiment code using :func:`log` and other methods
        #. Reading metrics and downloading artifacts when analyzing experimental results using :func:`get_metrics`
           and other methods

        To submit a run, create a configuration object that describes how the experiment is run.
        These objects include:

        * :class:`azureml.core.ScriptRunConfig`
        * :class:`azureml.train.automl.AutoMLConfig`
        * :class:`azureml.pipeline.core.PipelineRun`
        * :class:`azureml.pipeline.core.PublishedPipeline`


    :param experiment: The containing experiment.
    :type experiment: azureml.core.Experiment
    :param run_id: The run id for the run.
    :type run_id: str
    :param outputs: The outputs to be tracked.
    :type outputs: str
    :param _run_dto:
    :type _run_dto: azureml._restclient.models.run_dto.RunDto
    :param kwargs:
    """

    _RUNSOURCE_PROPERTY = BASE_RUN_SOURCE
    _run_source_initializers = {}

    def __init__(self, experiment, run_id, outputs=None, **kwargs):
        """Initialize the Run object.

        :param experiment: The containing experiment.
        :type experiment: azureml.core.Experiment
        :param run_id: The run id for the run.
        :type run_id: str
        :param outputs: The outputs to be tracked.
        :type outputs:
        :param kwargs:
        :type kwargs: dict
        """
        super(Run, self).__init__(experiment, run_id, outputs=outputs, **kwargs)
        self._container = self._run_dto.get("data_container_id", None)

    @staticmethod
    def _dto_to_run(experiment, run_dto, outputs=None, **kwargs):
        """Create a Run object from the service representation.

        :param experiment: The containing experiment.
        :type experiment: azureml.core.Experiment
        :param run_dto:
        :type run_dto: azureml._restclient.models.run_dto.RunDto
        :param outputs: Outputs directory to be tracked.
        :type outputs: str
        :param kwargs:
        :type kwargs: dict
        :return: Returns the Run object
        :rtype: Run
        """
        return Run(experiment, run_dto.run_id, outputs=outputs, _run_dto=run_dto, **kwargs)

    @classmethod
    def _load_scope(cls):
        """Load the current context from the environment.

        :return: experiment, run_id, url
        :rtype: azureml.core.Experiment, str, str
        """
        from .authentication import AzureMLTokenAuthentication
        from .experiment import Experiment
        from .workspace import Workspace

        try:
            # Load authentication scope environment variables
            subscription_id = os.environ['AZUREML_ARM_SUBSCRIPTION']
            run_id = os.environ["AZUREML_RUN_ID"]
            resource_group = os.environ["AZUREML_ARM_RESOURCEGROUP"]
            workspace_name = os.environ["AZUREML_ARM_WORKSPACE_NAME"]
            experiment_name = os.environ["AZUREML_ARM_PROJECT_NAME"]

            # Initialize an AMLToken auth, authorized for the current run
            token = os.environ['AZUREML_RUN_TOKEN']
            token_expiry_time = os.environ.get('AZUREML_RUN_TOKEN_EXPIRY',
                                               AzureMLTokenAuthentication._parse_expiry_time_from_token(token))

            url = os.environ["AZUREML_SERVICE_ENDPOINT"]
            location_from_url_regex_match = re.compile(r"//(.*?)\.").search(url)
            location = location_from_url_regex_match.group(1) if location_from_url_regex_match else None
        except KeyError as key_error:
            raise_from(RunEnvironmentException(), key_error)
        else:
            auth = AzureMLTokenAuthentication.create(token,
                                                     AzureMLTokenAuthentication._convert_to_datetime(
                                                         token_expiry_time),
                                                     url,
                                                     subscription_id,
                                                     resource_group,
                                                     workspace_name,
                                                     experiment_name,
                                                     run_id)
            # Disabling service check, as this is in remote context and we don't have an arm token
            # to check arm if the workspace exists or not.
            workspace = Workspace(subscription_id, resource_group, workspace_name,
                                  auth=auth, _location=location, _disable_service_check=True)

            experiment = Experiment(workspace, experiment_name, _skip_name_validation=True)

            return experiment, run_id

    @classmethod
    def get_context(cls, allow_offline=True, **kwargs):
        """Return current service context.

        Use this method to retrieve the current service context for logging metrics and uploading files. If
        allow_offline is set to True (and is by default), actions against the Run object will be printed to standard
        out.

        .. remarks::

            This function is commonly used to retrieve the authenticated Run object
            inside of a script to be submitted for execution via experiment.submit(). This run object is both
            an authenticated context to communicate with Azure ML services and a conceptual container within
            which metrics, files (artifacts), and models are contained.

            .. code-block:: python

                run = Run.get_context() # allow_offline=True by default, so can be run locally as well
                ...
                run.log("Accuracy", 0.98)
                run.log_row("Performance", epoch=e, error=err)

        :param cls:
        :type cls:
        :param allow_offline: Allow the service context to fall back to offline mode so that the training script
            can be tested locally without submitting a job with the SDK. True by default.
        :type allow_offline: bool
        :param kwargs:
        :type kwargs: dict
        :return: The submitted run.
        :rtype: azureml.core.run.Run
        """
        try:
            experiment, run_id = cls._load_scope()

            # Querying for the run instead of initializing to load current state
            return _SubmittedRun(experiment, run_id, **kwargs)
        except RunEnvironmentException as ex:
            module_logger.debug("Could not load run context %s, switching offline: %s", ex, allow_offline)
            if allow_offline:
                module_logger.info("Could not load the run context. Logging offline")
                return _OfflineRun(**kwargs)
            else:
                module_logger.debug("Could not load the run context and allow_offline set to False")
                raise RunEnvironmentException(inner_exception=ex)

    @classmethod
    def get_submitted_run(cls, **kwargs):
        """(Deprecated) use :func:`get_context`.

        Get the submitted run for this experiment.

        :return: The submitted run
        :rtype: azureml.core.run.Run
        """
        module_logger.warning("Run.get_submitted_run() is deprecated, use Run.get_context() instead")
        module_logger.warning("Run.get_submitted_run() will be removed shortly")
        return cls.get_context(allow_offline=False, **kwargs)

    @staticmethod
    def add_type_provider(runtype, run_factory):
        """Extensibility hook for custom Run types stored in Run History.

        :param runtype: The value of Run.type for which the factory will be invoked. Examples include
            'hyperdrive' or 'azureml.scriptrun', but can be extended with custom types.
        :type runtype: str
        :param run_factory: A function with signature (Experiment, RunDto) -> Run to be invoked when
            listing runs.
        :type run_factory: function
        """
        if Run._run_source_initializers.get(runtype, None) is not None:
            raise ValueError("RunType {0} already has a factory, can't add another".format(runtype))

        # TODO: Validate signature of the factory if possible via the code object
        module_logger.debug("Adding new factory {0} for run source {1}".format(run_factory, runtype))
        Run._run_source_initializers[runtype] = run_factory

    @staticmethod
    def _rehydrate_runs(experiment, run_dtos):
        """Rehydrate runs.

        :param experiment: The containing experiment.
        :param run_dtos:
        :type run_dtos: azureml._restclient.models.run_dto.RunDto
        """
        module_logger.debug("Available factories for run types {0}".format(Run._run_source_initializers))
        for run_dto in run_dtos:
            run_id = run_dto.run_id

            # TODO: Run source is around for backward compatibility. Delete after PuP
            run_properties = getattr(run_dto, "properties", {})
            run_source = run_properties.get(Run._RUNSOURCE_PROPERTY, 'None')
            runtype = run_dto.run_type if run_dto.run_type is not None else run_source

            module_logger.debug("Initializing Run {0} from type {1}".format(run_id, runtype))

            factory = Run._run_source_initializers.get(runtype, None)
            if factory is None and (runtype and not runtype.strip()):
                warnmsg = "Run {0} has type {1} but has no client type registered. "\
                          "Entire functionality might not be available."
                module_logger.warning(warnmsg.format(run_id, runtype))
            if factory is None:
                factory = Run._dto_to_run

            yield factory(experiment, run_dto)

    @staticmethod
    def list(
            experiment,
            type=None,
            tags=None,
            properties=None,
            status=None,
            include_children=False,
            _rehydrate_runs=True):
        """Get a list of runs in an experiment specified by optional filters.

        .. remarks::

            .. code-block:: python

                favorite_completed_runs = Run.list(experiment, status='Completed', tags='favorite')

                all_distinct_runs = Run.list(experiment)
                and_their_children = Run.list(experiment, include_children=True)

                only_script_runs = Run.list(experiment, type=ScriptRun.RUN_TYPE)

        :param experiment: The containing experiment.
        :type experiment: azureml.core.Experiment
        :param type: If specified, returns runs matching specified type.
        :type type: str
        :param tags:  If specified, returns runs matching specified *"tag"* or {*"tag"*: *"value"*}.
        :type tags: str or dict
        :param properties: If specified, returns runs matching specified *"property"* or {*"property"*: *"value"*}.
        :type properties: str or dict
        :param status: If specified, returns runs with status specified *"status"*.
        :type status: str
        :param include_children: If set to true, fetch all the runs, not only top-level ones.
        :type include_children: bool
        :param _rehydrate_runs: If set to True (by default), will use the registered provider to reinstantiate
            an object for that type instead of the base Run.
        :type _rehydrate_runs: bool
        :return: list of azureml.core.run.Run
        :rtype: :py:class:`list[azureml.core.run.Run]`
        """
        from azureml._run_impl.run_history_facade import RunHistoryFacade

        run_dtos = RunHistoryFacade.get_runs(
            experiment,
            runtype=type,
            tags=tags,
            properties=properties,
            status=status,
            include_children=include_children)

        module_logger.debug("Rehydrating runs on enumerate: {0}".format(_rehydrate_runs))
        if _rehydrate_runs:
            return Run._rehydrate_runs(experiment, run_dtos)
        else:
            return (Run._dto_to_run(experiment, run_dto) for run_dto in run_dtos)

    @staticmethod
    def _start_logging(experiment, name=None, run_id=None,
                       outputs=None, snapshot_directory=".", **kwargs):
        """Create and start a new run for the workspace and experiment name.

        :param experiment: The containing experiment.
        :type experiment: azureml.core.Experiment
        :param name: Optional name for the run.
        :type name: str
        :param run_id: Optional run_id for the run, otherwise uses default.
        :type run_id: str
        :param outputs: Optional outputs directory to track.
        :type outputs: str
        :param snapshot_directory: Optional directory to take snapshot of. Setting to None will take no snapshot.
        :type snapshot_directory: str
        :return: Return a run.
        :rtype: azureml.core.run.Run
        :param kwargs:
        """
        outputs = outputs if outputs else OUTPUTS_DIR
        run = Run._create(experiment, name=name,
                          run_id=run_id, outputs=outputs, **kwargs)
        run._client.start()
        if snapshot_directory:
            run._client.take_snapshot(snapshot_directory, _raise_on_validation_failure=False)
        return run

    @staticmethod
    def _create(experiment, name=None, run_id=None,
                outputs=None, **kwargs):
        """Create a new run for the workspace and experiment name.

        :param experiment: The containing experiment.
        :type experiment: azureml.core.Experiment
        :param name: Optional name for the run.
        :type name: str
        :param run_id: Optional run_id for the run, otherwise uses default.
        :type run_id: str
        :param outputs: Optional outputs directory to track.
        :type outputs: str
        :return: Return a run.
        :rtype: azureml.core.run.Run
        :param kwargs:
        """
        from azureml._run_impl.run_history_facade import RunHistoryFacade
        run_dto = RunHistoryFacade.create_run(experiment, name=name, run_id=run_id)
        return Run._dto_to_run(experiment, run_dto, outputs=outputs, **kwargs)

    @property
    def id(self):
        """Get Run id.

        The id of the run - an identifier unique across the containing experiment

        :return: The run id
        :rtype: str
        """
        return self._run_id

    @property
    def name(self):
        """Return the Run name.

        The optional name of the run - a user-set name for later identification

        :return: The run id
        :rtype: str
        """
        return self._run_name

    @property
    def type(self):
        """Get Run type.

        Indicates how the run was created or configured.

        :return: The run type
        :rtype: str
        """
        return self._runtype

    @property
    def number(self):
        """Get Run number.

        A monotonically increasing number representing the order of runs within an experiment.

        :return: The run number.
        :rtype: int
        """
        self._logger.warning("Run Number's semantics may be modified in the future")
        return self._run_number

    @property
    def parent(self):
        """Fetch the parent run for this run from the service.

        Runs can have an optional parent, resulting in a potential tree hierarchy of Runs.

        :return: The parent Run, or None if one is not set
        :rtype: azureml.core.run.Run
        """
        parent_run_id = self._run_dto.get('parent_run_id', None)
        return None if parent_run_id is None else get_run(self.experiment, parent_run_id)

    @property
    def status(self):
        """Return the run object's status."""
        return self._latest_status

    def get_status(self):
        """Fetch the run's latest status from the service.

        Common values returned include "Running", "Completed", and "Failed".

        .. remarks::

            * NotStarted - This is a temporary state client-side Run objects are in before cloud submission
            * Starting - The Run has started being processed in the cloud - the caller has a run ID at this point
            * Provisioning - Used when on-demand compute is being created for a given job submission
            * Preparing - The run environment is being prepared:
                * docker image build
                * conda enviornment setup
            * Queued - The job is queued in the compute target. For example, in BatchAI the job is in queued state
                 while waiting for all the requested nodes to be ready.
            * Running - The job started to run in the compute target.
            * Finalizing - User code has completed and the run is in post-processing stages.
            * CancelRequested - Cancellation has been requested for the job.
            * Completed - The run completed successfully. This includes both the user code and run
                post-processing stages.
            * Failed - The run failed. Usually the Error property on a run will provide details as to why.
            * Canceled - Following cancellation request, the run is now successfully cancelled.
            * NotResponding - For runs that have Heartbeats enabled, no heartbeat has been recently sent.

            .. code-block:: python

                run = experiment.submit(config)
                while run.get_status() not in ['Completed', 'Failed']: # For example purposes only, not exhaustive
                    print('Run {} not in terminal state'.format(run.id))
                    time.sleep(10)

        :return: The latest status as a string
        :rtype: str
        """
        if self._latest_status not in ["Completed", "Failed", "Canceled"]:
            self._latest_status = self._client.get_status()
        return self._latest_status

    def get_details(self):
        """Get the definition, status information, current log files and other details of the run.

        .. remarks::

            The returned dictionary contains the following key-value pairs:

            * `runId`: ID of this run.
            * `target`
            * `status`: The run's current status. Same value as that returned from `get_status()`.
            * `startTimeUtc`: UTC time of when this run was started, in ISO8601.
            * `endTimeUtc`: UTC time of when this run was finished (either Completed or Failed), in ISO8601.
                This key does not exist if the run is still in progress.
            * `properties`
            * `logFiles`

        :return: Return the details for the run
        :rtype: dict[str, str]
        """
        return self._client.get_runstatus().as_dict(key_transformer=camel_case_transformer)

    def get_children(self, recursive=False, tags=None, properties=None, type=None, status=None, _rehydrate_runs=True):
        """Get all children for the current run selected by specified filters.

        :param rehydrate_runs: Indicates whether to construct the original Run class or the base Run class
        :type rehydrate_runs: bool
        :param recursive: Indicates whether to recurse through all descendants.
        :type recursive: bool
        :param type: If specified, returns runs matching specified type.
        :type type: str
        :param tags:  If specified, returns runs matching specified *"tag"* or {*"tag"*: *"value"*}.
        :type tags: str or dict
        :param properties: If specified, returns runs matching specified *"property"* or {*"property"*: *"value"*}.
        :type properties: str or dict
        :param status: If specified, returns runs with status specified *"status"*.
        :type status: str
        :param _rehydrate_runs: Indicates whether to instantiate a run of the original type or the base Run.
        :type _rehydrate_runs: bool
        :return: A list of azureml.core.run.Run
        :rtype: :class:`list`
        """
        children = self._client.get_descendants(
            root_run_id=self._root_run_id,
            recursive=recursive,
            tags=tags,
            properties=properties,
            runtype=type,
            status=status)

        self._logger.debug("Rehydrating runs on enumerate: {0}".format(_rehydrate_runs))
        if _rehydrate_runs:
            return Run._rehydrate_runs(self.experiment, children)
        else:
            return (Run._dto_to_run(self.experiment, child) for child in children)

    def get_metrics(self, name=None, recursive=False, run_type=None, populate=False):
        """Retrieve the metrics logged to the run.

        If recursive=True (False by default), fetch metrics for runs in the given run's subtree.

        .. remarks::

            .. code-block:: python

                run = experiment.start_logging() # run id: 123
                run.log("A", 1)
                with run.child_run() as child: # run id: 456
                    child.log("A", 2)

                metrics = run.get_metrics()
                # metrics = { 'A': 1 }

                metrics = run.get_metrics(recursive=True)
                # metrics = { '123': { 'A': 1 }, '456': { 'A': 2 } } note key is runId

        :param name: The name of the metric.
        :type name: str
        :param recursive: Indicates whether to recurse through all descendants.
        :type recursive: bool
        :param run_type:
        :type run_type: str
        :param populate: When True (False by default), fetch the contents of external data linked to the metric
        :type populate: bool
        :return: A dictionary containing the users metrics.
        :rtype: dict
        """
        if name is not None:
            raise NotImplementedError("Cannot filter on metric name yet")

        if recursive:
            # TODO: No better way?
            descendant_ids = [
                child.run_id for child in self._client.get_descendants(
                    root_run_id=self._root_run_id,
                    recursive=True,
                    runtype=run_type)
            ]
            metrics = self._client.run.get_metrics_by_run_ids(run_ids=descendant_ids + [self.id])
            # TODO: changing return structure based on parameter is awkward
            # TODO: Refactor metrics client to be a bit more composable
            metrics_dtos_by_run = defaultdict(list)
            for dto in metrics:
                metrics_dtos_by_run[dto.run_id].append(dto)
            ret_metrics = {
                runid: self._client.metrics.dto_to_metrics_dict(metric_dto_list)
                for runid, metric_dto_list in metrics_dtos_by_run.items()
            }
        else:
            ret_metrics = self._client.metrics.get_all()

        if populate:
            artifacts = self.get_file_names()
            for name in ret_metrics:
                if ret_metrics[name] is None:
                    if name in artifacts:
                        artifact_contents = self._download_artifact_contents_to_string(name)
                        if artifact_contents is None:
                            self._logger.debug("Unable to retrieve metric artifact: {}".format(name))
                            ret_metrics[name] = None
                        else:
                            try:
                                ret_metrics[name] = json.loads(artifact_contents)
                            except json.decoder.JSONDecodeError as e:
                                self._logger.debug("Unable to decode metric artifact: {}".format(e))
                                ret_metrics[name] = None
        return ret_metrics

    @property
    def experiment(self):
        """Get experiment containing the run.

        :return: Retrieves the experiment corresponding to the run.
        :rtype: azureml.core.Experiment
        """
        return self._experiment

    def start(self):
        """Mark the run as started.

        This is typically used in advanced scenarios when the run has been created by another actor.
        """
        self._client.start()

    def complete(self, _set_status=True):
        """Wait for task queue to be processed.

        Then run is marked as completed. This is typically used in interactive notebook scenarios.

        :param _set_status: send the status event for tracking
        :type _set_status: bool
        """
        self._client.complete(_set_status=_set_status)

    def fail(self, error_details=None, _set_status=True):
        """Mark the run as failed.

        Optionally set the Error property of the run with a message or exception passed to error_details.

        :param error_details: Optional details of the error
        :type error_details: Instance of string or BaseException
        :param _set_status: send the status event for tracking
        :type _set_status: bool
        """
        self._client.fail(error_details=error_details, _set_status=_set_status)

    def cancel(self):
        """Mark the run as canceled."""
        self._client.cancel()

    def __enter__(self):
        """Initialize features related to ouputs tracking and status updates for the run.".

        :return:
        :rtype: Run
        """
        super(Run, self).__enter__()
        return self

    def __exit__(self, exit_type, value, traceback):
        """Close tracking and collect outputs.

        :param exit_type:
        :type exit_type:
        :param value:
        :type value:
        :param traceback:
        :type traceback:
        :return:
        :rtype: Run
        """
        return super(Run, self).__exit__(exit_type, value, traceback)

    def child_run(self, name=None, run_id=None, outputs=None):
        """Create a child run.

        .. remarks::

            This is used to isolate part of a run into a subsection.  This can be done for
            identifiable "parts" of a run that are interesting to seperate, or to capture
            independent metrics across an interation of a subprocess.

            If an output directory is set for the child run, the contents of that directory will be
            uploaded to the child run record when the child is completed.

        :param name: Optional name for the child, typically specified for a "part"
        :type name: str
        :param run_id: Optional run_id for the child, otherwise auto-generated.  Typically not set
        :type run_id: str
        :param outputs: Optional outputs directory to track for the child
        :type outputs: str
        :return: The child run
        :rtype: azureml.core.run.Run
        """
        run_dto = self._run_dto
        child_dto = self._client.create_child_run(run_dto.get("name"),
                                                  run_dto.get("target", SDK_TARGET),
                                                  child_name=name,
                                                  run_id=run_id)
        child = Run._dto_to_run(self.experiment, child_dto, outputs=outputs,
                                _worker_pool=self._client.worker_pool, _parent_logger=self._logger)
        child._client.start()
        return child

    def create_children(self, count=None, tag_key=None, tag_values=None):
        """Create one or many child runs.

        .. remarks::

            Either count OR tag_key AND tag_values must be specified

        :param count: Optional number of children to create
        :type count: int
        :param tag_key: Optional key for the Tags entry to populate in all created children
        :type tag_key: str
        :param tag_Values: Optional list of values that will map onto Tags[tag_key] for the list of runs created
        :type tag_values: [str]
        :return: The list of child runs
        :rtype: [azureml.core.run.Run]
        """
        if tag_key is None and tag_values is None and count is None:
            raise UserErrorException("Either count or tag_key and tag_values must be set")
        if tag_key is not None or tag_values is not None:
            if tag_key is None:
                raise UserErrorException("tag_key must not be null when tag_values is set")
            if tag_values is None:
                raise UserErrorException("tag_values must not be null when tag_key is set")
            if count is not None:
                raise UserErrorException("count can not be set when tag_key and tag_values are set")
            if len(tag_values) > 100:
                raise UserErrorException("create_children can not create more than 100 runs")

        if count is not None:
            if count > 100:
                raise UserErrorException("create_children can not create more than 100 runs")
            tag_key = "azureml.batch_child_number"
            tag_values = list(range(count))

        child_dtos = self._client.create_children(tag_key, tag_values)
        children = []
        for child_dto in child_dtos:
            child = Run._dto_to_run(self.experiment, child_dto,
                                    _worker_pool=self._client.worker_pool, _parent_logger=self._logger)
            child._client.start()
            children.append(child)
        return children

    def tag(self, key, value=None):
        """Tag the run with a string key and optional string value.

        .. remarks::

            Tags and Properties on a run are both dictionaries of string -> string. The difference between them
            is mutability: Tags can be set, updated, and deleted while Properties can only be added. This makes
            Properties more appropriate for system/workflow related behavior triggers, while Tags are generally
            user-facing and meaningful for the consumers of the experiment.

            .. code-block:: python

                run = experiment.start_logging()
                run.tag('DeploymentCandidate')
                run.tag('modifiedBy', 'Master CI')
                run.tag('modifiedBy', 'release pipeline') # Careful, tags are mutable

                run.add_properties({'BuildId': os.environ.get('VSTS_BUILD_ID')}) # Properties are not

                tags = run.get_tags()
                # tags = { 'DeploymentCandidate': None, 'modifiedBy': 'release pipeline' }


        :param key: The tag key
        :type key: str
        :param value: An optional value for the tag
        :type value: str
        """
        self._client.set_tag(key, value)

    def set_tags(self, tags):
        """Add or modify a set of tags on the run.

        :param tags: The tags stored in the run object
        :type tags: dict[str] = str
        """
        self._client.set_tags(tags)

    @property
    def tags(self):
        """Return the set of mutable tags on this run.

        :return: The tags stored on the run object
        :rtype: dict
        """
        return self._run_dto.get('tags', {})

    def get_tags(self):
        """Fetch the latest set of mutable tags on the run from the service.

        :return: The tags stored on the run object
        :rtype: dict
        """
        return self._client.get_tags()

    def add_properties(self, properties):
        """Add immutable properties to the run. Tags and Properties (both dict[str, str]) differ in their mutability.

        :param properties: The hidden properties stored in the run object
        :type properties: dict
        """
        self._client.add_properties(properties)

    @property
    def properties(self):
        """Return the immutable properties of this run.

        .. remarks::

            Properties include immutable system-generated information such as
            duration, date of execution, user, etc.

        :return: The locally cached properties of the run
        :rtype: dict[str] = str
        """
        return self._run_dto.get('properties', {})

    def get_properties(self):
        """Fetch the latest properties of the run from the service.

        .. remarks::

            Properties include immutable system-generated information such as
            duration, date of execution, user, etc.

        :return: The properties of the run
        :rtype: dict
        """
        return self._client.get_properties()

    def log(self, name, value, description=""):
        """Log a metric value to the run with the given name.

        .. remarks::

            Logging a metric to a run causes that metric to be stored in
            the run record in the experiment.  You can log the same metric
            multiple times within a run, the result being considered a vector
            of that metric.

            .. code-block: python

                run = experiment.start_logging()
                run.log('Accuracy', (tp + tn) / (p + n))
                run.log('Precision', tp / (tp + fp))
                run.log('Recall', tp / (tp + fn))

        :param name: The name of metric
        :type name: str
        :param value: The value to be posted to the service
        :type value:
        :param description: An optional metric description
        :type description: str
        """
        # python2/3 compatible check for range, range is not a python2 type
        is_range = isinstance(value, type(range(0)))
        if any([isinstance(value, list),
                is_range,
                isgeneratorfunction(value),
                isinstance(value, GeneratorType)]):
            raise UserErrorException("run.log does not support "
                                     "list type. Use run.log_list instead")
        else:
            self._client.log_scalar(name, value, description=description)

    def log_list(self, name, value, description=""):
        """Log a list metric value to the run with the given name.

        :param name: The name of metric
        :type name: str
        :param value: The name of metric
        :type value: :class:`list`
        :param description: An optional metric description
        :type description: str
        """
        self._client.log_list(name, value, description=description)

    def log_row(self, name, description=None, **kwargs):
        """Log a row metric to the run with the given name.

        .. remarks::

            Using *log_row* creates a table metric with columns as described in kwargs.  Each named
            parameter generates a column with the value specified.  *log_row* can be
            be called once to log an aribitrary tuple, or multiple times in a loop to generate
            a complete table.

            .. code-block:: python

                citrus = ['orange', 'lemon', 'lime']
                sizes = [ 10, 7, 3]
                for index in range(len(citrus)):
                    run.log_row("citrus", fruit = citrus[index], size=sizes[index])

        :param name: The name of metric
        :type name: str
        :param description: An optional metric description
        :type description: str
        :param kwargs:
        """
        self._client.log_row(name, kwargs, description=description)

    def log_table(self, name, value, description=""):
        """Log a table metric to the run with the given name.

        :param name: The name of metric
        :type name: str
        :param value: The table value of the metric
        :type value: str
        :type value dict: The dictionary of str to columns to be posted to the service
        :param description: An optional metric description
        :type description: str
        """
        self._client.log_table(name, value, description=description)

    @_check_for_data_container_id
    def log_confusion_matrix(self, name, value, description=""):
        """Log a confusion matrix to the artifact store.

        This logs a wrapper around the sklearn confusion matrix.
        The metric data contains the class labels and a 2D list
        for the matrix itself.
        See the following link for more details on how the metric is computed:
        `https://scikit-learn.org/stable/modules/generated/sklearn.metrics.confusion_matrix.html`

        .. remarks::

            Example of a valid JSON value:

            .. code-block:: json

                {
                    schema_type: "confusion_matrix",
                    schema_version: "v1",
                    "data": {
                        "class_labels": ["0", "1", "2", "3"],
                        "matrix": [
                            [3, 0, 1, 0],
                            [0, 1, 0, 1],
                            [0, 0, 1, 0],
                            [0, 0, 0, 1]
                        ]
                    }
                }

        :param name: The name of the confusion matrix
        :type name: str
        :param value: json containing name, version, and data properties
        :type value: str
        :param description: An optional metric description
        :type description: str
        :return:
        """
        self._client.log_confusion_matrix(name, value, description=description)

    @_check_for_data_container_id
    def log_accuracy_table(self, name, value, description=""):
        """Log a accuracy table to the artifact store.

        The accuracy table metric is a multi-use non-scalar metric
        that can be used to produce multiple types of line charts
        that vary continuously over the space of predicted probabilities.
        Examples of these charts are ROC, precision-recall, and lift curves.

        The calculation of the accuracy table is similar to the calculation
        of an ROC curve. An ROC curve stores true positive rates and
        false positive rates at many different probability thresholds.
        The accuracy table stores the raw number of
        true positives, false positives, true negatives, and false negatives
        at many probability thresholds.

        There are two methods used for selecting thresholds: "percentile" and
        "percentile." They differ in how they sample from the space of
        predicted probabilities.

        Probability thresholds are uniformly spaced thresholds between 0 and 1.
        If NUM_POINTS were 5 the probability thresholds would be
        [0.0, 0.25, 0.5, 0.75, 1.0].

        Percentile thresholds are spaced according to the distribution of
        predicted probabilities. Each threshold corresponds to the percentile
        of the data at a probability threshold.
        For example, if NUM_POINTS were 5, then the first threshold would be at
        the 0th percentile, the second at the 25th percentile, the
        third at the 50th, and so on.

        The probability tables and percentile tables are both 3D lists where
        the first dimension represents the class label*, the second dimension
        represents the sample at one threshold (scales with NUM_POINTS),
        and the third dimension always has 4 values: TP, FP, TN, FN, and
        always in that order.

        * The confusion values (TP, FP, TN, FN) are computed with the
        one vs. rest strategy. See the following link for more details:
        `https://en.wikipedia.org/wiki/Multiclass_classification`

        .. remarks::

            Example of a valid JSON value:

            .. code-block:: json

                {
                    "schema_type": "accuracy_table",
                    "schema_version": "v1",
                    "data": {
                        "probability_tables": [
                            [
                                [82, 118, 0, 0],
                                [75, 31, 87, 7],
                                [66, 9, 109, 16],
                                [46, 2, 116, 36],
                                [0, 0, 118, 82]
                            ],
                            [
                                [60, 140, 0, 0],
                                [56, 20, 120, 4],
                                [47, 4, 136, 13],
                                [28, 0, 140, 32],
                                [0, 0, 140, 60]
                            ],
                            [
                                [58, 142, 0, 0],
                                [53, 29, 113, 5],
                                [40, 10, 132, 18],
                                [24, 1, 141, 34],
                                [0, 0, 142, 58]
                            ]
                        ],
                        "percentile_tables": [
                            [
                                [82, 118, 0, 0],
                                [82, 67, 51, 0],
                                [75, 26, 92, 7],
                                [48, 3, 115, 34],
                                [3, 0, 118, 79]
                            ],
                            [
                                [60, 140, 0, 0],
                                [60, 89, 51, 0],
                                [60, 41, 99, 0],
                                [46, 5, 135, 14],
                                [3, 0, 140, 57]
                            ],
                            [
                                [58, 142, 0, 0],
                                [56, 93, 49, 2],
                                [54, 47, 95, 4],
                                [41, 10, 132, 17],
                                [3, 0, 142, 55]
                            ]
                        ],
                        "probability_thresholds": [0.0, 0.25, 0.5, 0.75, 1.0],
                        "percentile_thresholds": [0.0, 25.0, 50.0, 75.0, 100.0],
                        "class_labels": ["0", "1", "2"]
                    }
                }

        N = # of samples in validation dataset (200 in example)
        M = # thresholds = # samples taken from the probability space (5 in example)
        C = # classes in full dataset (3 in example)

        Some invariants of the accuracy table:
        - TP + FP + TN + FN = N for all thresholds for all classes
        - TP + FN is the same at all thresholds for any class
        - TN + FP is the same at all thresholds for any class
        - Probability tables and percentile tables have shape [C, M, 4]

        Note: M can be any value and controls the resolution of the charts
        This is independent of the dataset, is defined when calculating
        metrics, and trades off storage space, computation time, and resolution.

        Class labels should be strings, confusion values should be integers, and
        thresholds should be floats.

        :param name: The name of the accuracy table
        :type name: str
        :param value: json containing name, version, and data properties
        :type value: str
        :param description: An optional metric description
        :type description: str
        :return:
        """
        self._client.log_accuracy_table(name, value, description=description)

    @_check_for_data_container_id
    def log_residuals(self, name, value, description=""):
        """Log a residuals to the artifact store.

        This logs the data needed to display a histogram of
        residuals for a regression task.
        The residuals are predicted - actual.

        There should be one more edge than the number of counts.
        Please see the numpy histogram documentation for examples of
        using counts and edges to represent a histogram.
        `https://docs.scipy.org/doc/numpy/reference/generated/numpy.histogram.html`

        .. remarks::

            Example of a valid JSON value:

            .. code-block:: json

                {
                    "schema_type": "residuals",
                    "schema_version": "v1",
                    "data": {
                        "bin_edges": [50, 100, 200, 300, 350],
                        "bin_counts": [0.88, 20, 30, 50.99]
                    }
                }

        :param name: The name of the residuals
        :type name: str
        :param value: json containing name, version, and data properties
        :type value: str
        :param description: An optional metric description
        :type description: str
        :return:
        """
        self._client.log_residuals(name, value, description=description)

    @_check_for_data_container_id
    def log_predictions(self, name, value, description=""):
        """Log a predictions to the artifact store.

        This logs a metric score that can be used to compare
        the distributions of true target values to the
        distribution of predicted values for a regression task.

        The predictions are binned and standard deviations are calculated
        for error bars on a line chart.

        .. remarks::

            Example of a valid JSON value:

            .. code-block:: json

                {
                    "schema_type": "predictions",
                    "schema_version": "v1",
                    "data": {
                        "bin_averages": [0.25, 0.75],
                        "bin_errors": [0.013, 0.042],
                        "bin_counts": [56, 34],
                        "bin_edges": [0.0, .5, 1.0]
                    }
                }

        :param name: The name of the predictions
        :type name: str
        :param value: json containing name, version, and data properties
        :type value: str
        :param description: An optional metric description
        :type description: str
        :return:
        """
        self._client.log_predictions(name, value, description=description)

    @_check_for_data_container_id
    def log_image(self, name, path=None, plot=None):
        """Log an image metric to the run record.

        .. remarks::

            Use log_image to log an image file or a matplotlib plot to the run. These images will
            be visible and comparable in the run record.

        :param name: The name of metric
        :type name: str
        :param path: The path or stream of the image
        :type path: str
        :param plot: The plot to log as an image
        :type plot: matplotlib.pyplot
        """
        return self._client.log_image(name, path=path, plot=plot)

    @_check_for_data_container_id
    def upload_file(self, name, path_or_stream):
        """Upload a file to the run record.

        .. remarks::

            .. code-block: python

                run = experiment.start_logging()
                run.upload_file(name='important_file', path="path/on/disk/file.txt")

                .. note::

                    Runs automatically capture file in the specified output directory, which defaults
                    to "./outputs" for most run types.  Use upload_file only when additional files need
                    to be uploaded or an output directory is not specified.

        :param name: The name of the file to upload.
        :type name: str
        :param path_or_stream: The relative local path or stream to the file to upload.
        :type path_or_stream: str
        :return:
        :rtype: azure.storage.blob.models.ResourceProperties
        """
        return self._client.artifacts.upload_artifact(path_or_stream, RUN_ORIGIN, self._container, name)

    @_check_for_data_container_id
    def upload_folder(self, name, path):
        """Upload the specified folder to the given prefix name.

        .. remarks::

            .. code-block: python

                run = experiment.start_logging()
                run.upload_folder(name='important_files', path='path/on/disk')

                run.download_file('important_files/existing_file.txt', 'local_file.txt')

                .. note::

                    Runs automatically capture files in the specified output directory, which defaults
                    to "./outputs" for most run types.  Use upload_folder only when additional files need
                    to be uploaded or an output directory is not specified.

        :param name: The name of the folder of files to upload.
        :type name: str
        :param folder: The relative local path to the folder to upload.
        :type folder: str
        """
        return self._client.artifacts.upload_dir(path, RUN_ORIGIN, self._container,
                                                 lambda fpath: name + fpath[len(path):])

    @_check_for_data_container_id
    def upload_files(self, names, paths):
        """Upload files to the run record.

        .. remarks::

            *upload_files* has the same effect as upload_file on separate files, however
            there are performance and resource utilization benefits when using *upload_files*.

            .. code-block: python

                import os

                run = experiment.start_logging()
                file_name_1 = 'important_file_1'
                file_name_2 = 'important_file_2'
                run.upload_files(names=[file_name_1, file_name_2],
                                 paths=['path/on/disk/file_1.txt', 'other/path/on/disk/file_2.txt'])

                run.download_file(file_name_1, 'file_1.txt')

                os.mkdir("path")  # The path must exist
                run.download_file(file_name_2, 'path/file_2.txt')

                .. note::

                    Runs automatically capture files in the specified output directory, which defaults
                    to "./outputs" for most run types.  Use upload_files only when additional files need
                    to be uploaded or an output directory is not specified.

        :param names: The names of the files to upload. If set, paths must also be set.
        :type names: :class:`list`
        :param paths: The relative local paths to the files to upload. If set, names is required.
        :type paths: :class:`list`
        """
        if len(names) != len(paths):
            raise TypeError("names and paths must be lists of equal lengths. "
                            "Names was of length {} and paths of length {}.".format(len(names), len(paths)))
        return self._client.artifacts.upload_files(paths, RUN_ORIGIN, self._container, names)

    def get_file_names(self):
        """List the files that are stored in association with the run.

        :return: The list of paths for existing artifacts
        :rtype: :class:`list[str]`
        """
        if self._container is None:
            return []
        else:
            return list(self._client.artifacts.get_file_paths(RUN_ORIGIN, self._container))

    @_check_for_data_container_id
    def download_file(self, name, output_file_path=None):
        """Download an associated file from storage.

        :param name: The name of the artifact to be downloaded
        :type name: str
        :param output_file_path: The local path where to store the artifact
        :type output_file_path: str
        """
        if output_file_path is None:
            self._logger.debug("output_file_path for download_file was not set.")
            output_file_path = os.path.basename(name)

        self._client.artifacts.download_artifact(RUN_ORIGIN, self._container, name, output_file_path)

    @_check_for_data_container_id
    def download_files(self, prefix=None, output_directory=None, output_paths=None, batch_size=100):
        """Download files from a given storage prefix (folder name) or the entire container if prefix is unspecified.

        :param prefix: the filepath prefix within the container from which to download all artifacts
        :type prefix: str
        :param output_directory: optional directory that all artifact paths use as a prefix
        :type output_directory: str
        :param output_paths: optional filepaths in which to store the downloaded artifacts.
            Should be unique and match length of paths.
        :type output_paths: [str]
        :param batch_size: number of files to download per batch
        :type batch_size: int
        """
        self._client.artifacts.download_artifacts_from_prefix(RUN_ORIGIN,
                                                              self._container,
                                                              prefix=prefix,
                                                              output_directory=output_directory,
                                                              output_paths=output_paths,
                                                              batch_size=batch_size)

    @_check_for_data_container_id
    def _download_artifact_contents_to_string(self, name):
        """Download the contents of an artifact saved to the run."""
        return self._client.artifacts.download_artifact_contents_to_string(
            origin=RUN_ORIGIN, container=self._container, path=name)

    def _take_snapshot(self, file_or_folder_path, _raise_on_validation_failure=True):
        """Save a snapshot of the inputted file or folder.

        :param file_or_folder_path: The file or folder containing the run source code.
        :type file_or_folder_path: str
        :return: Returns the snapshot id
        :param _raise_on_validation_failure: If set to True (by default), will raise an exception on validation errors
        :type _raise_on_validation_failure: bool
        :rtype: str
        """
        return self._client.take_snapshot(file_or_folder_path)

    def take_snapshot(self, file_or_folder_path):
        """Save a snapshot of the inputted file or folder.

        .. remarks::

            Snapshots are intended to be the *source code* used to execute the experiment run.
            These are stored with the run so that the run trial can be replicated in the future.

            .. note::

                Snapshots are automatically taken when :func:azureml.core.Experiment.submit is called.
                Typically take_snapshot is only required for interactive (notebook) runs.

        :param file_or_folder_path: The file or folder containing the run source code.
        :type file_or_folder_path: str
        :return: Returns the snapshot id
        :rtype: str
        """
        return self._take_snapshot(file_or_folder_path)

    def restore_snapshot(self, snapshot_id=None, path=None):
        """Restores a snapshot as a zip file. Returns the path to the zip.

        :param snapshot_id: The snapshot id to restore, the latest if not specified.
        :type snapshot_id: str
        :param path: The path where the downloaded zip is saved.
        :type path: str
        :return: The path.
        :rtype: str
        """
        if snapshot_id is None:
            snapshot_id = self.get_snapshot_id()

        return self._client.snapshots.restore_snapshot(snapshot_id, path)

    def get_snapshot_id(self):
        """Get the latest snapshot id.

        :return: The most recent snapshot id.
        :rtype: str
        """
        return self._client.get_snapshot_id()

    def register_model(self, model_name, model_path=None, tags=None, properties=None, **kwargs):
        """Register a model for operationalization.

        .. remarks::

            .. code-block:: python

                model = best_run.register_model(model_name = 'best_model', model_path = 'outputs/model.pkl')

        :param model_name:
        :type model_name: str
        :param model_path: relative cloud path to model from outputs/ dir. When model_path is None, model_name is path.
        :type model_path: str
        :param tags: Dictionary of key value tags to give the model
        :type tags: dict[str, str]
        :param properties: Dictionary of key value properties to give the model. These properties cannot
            be changed after model creation, however new key value pairs can be added
        :type properties: dict[str, str]
        :param kwargs:
        :return:
        :rtype: azureml.core.model.Model
        """
        return self._client.register_model(model_name, model_path, tags, properties, **kwargs)


def get_run(experiment, run_id, rehydrate=True):
    """Get the run for this experiment with its run id.

    :param experiment: The containing experiment.
    :type experiment: azureml.core.Experiment
    :param run_id: The run id for the run.
    :type run_id: string
    :param rehydrate: flag indicates if it is a rehydrate run
    :type rehydrate: boolean
    :return: The submitted run.
    :rtype: azureml.core.run.Run
    """
    from azureml._restclient.run_client import RunClient
    client = RunClient(experiment.workspace.service_context, experiment.name, run_id)
    run_dto = client.get_run()

    module_logger.debug("Rehydrating run: {0}".format(rehydrate))
    if rehydrate:
        run_properties = getattr(run_dto, "properties", {})
        run_source = run_properties.get(Run._RUNSOURCE_PROPERTY, 'None')
        runtype = run_dto.run_type if run_dto.run_type is not None else run_source
        if runtype in Run._run_source_initializers:
            factory = Run._run_source_initializers[runtype]
            return factory(experiment, run_dto)

        module_logger.info("Run type {} is set but does not have a hydration factory".format(runtype))
    return Run._dto_to_run(experiment, run_dto)


class _SubmittedRun(Run):
    """Run class for remote execution."""

    def complete(self):
        """Override complete since it is not supported for submitted runs."""
        self._logger.info("complete is not supported for submitted runs.")
        super(_SubmittedRun, self).complete(_set_status=False)

    def fail(self, error_details=None):
        """Override fail since it is not supported for submitted runs."""
        self._logger.info("fail is not setting status for submitted runs.")
        super(_SubmittedRun, self).fail(error_details=error_details, _set_status=False)

    def cancel(self):
        """Override cancel since it is not supported for submitted runs."""
        self._logger.warning("cancel is not supported for submitted runs.")

    def __enter__(self):
        """Override __enter__ since it is not supported for submitted runs."""
        self._logger.warning("Submitted runs are not supported context managers.")

    def __exit__(self, *args):
        """Override __exit__ since it is not supported for submitted runs.

        :param args:
        """
        self._logger.warning("Submitted runs are not supported context managers.")


class _OfflineRun(ChainedIdentity):
    def __init__(self, parent_logger=None, run_id=None, **kwargs):

        self._run_id = "OfflineRun_{}".format(uuid4()) if run_id is None else run_id
        super(_OfflineRun, self).__init__(
            _ident=self._run_id,
            _parent_logger=parent_logger if parent_logger is not None else module_logger)

    @property
    def id(self):
        return self._run_id

    def complete(self):
        self._logger.debug("complete is not supported for offline runs.")

    def fail(self, error_details=None):
        self._logger.debug("fail is not supported for offline runs.")
        print("Run failed:\n{}".format(str(error_details)))

    def cancel(self):
        self._logger.debug("cancel is not supported for offline runs.")

    def log(self, name, value, description=""):
        self._emit("scalar", name, value)

    def log_list(self, name, value, description=""):
        self._emit("list", name, value)

    def log_table(self, name, value, description=""):
        self._emit("table", name, value)

    def log_row(self, name, description=None, **kwargs):
        self._emit("row", name, kwargs)

    def log_image(self, name, path=None, plot=None):
        self._emit("image", name, plot or path)

    def upload_file(self, name, path_or_stream):
        print("Attempted to track file {0} at {1}".format(name, path_or_stream))

    def child_run(self, name=None, run_id=None, outputs=None):
        childname = str(uuid4()) if name is None else name
        newrid = "{}.{}".format(self.id, childname) if run_id is None else run_id
        return _OfflineRun(parent_logger=self._logger, run_id=newrid)

    def __enter__(self):
        print("[START] Run {} context".format(self.id))
        return self

    def __exit__(self, exit_type, value, traceback):
        print("[STOP] Run {} context".format(self.id))

    def _emit(self, type, name, value):
        print("Attempted to log {0} metric {1}:\n{2}".format(type, name, value))
